/***************************************************************************

Copyright (c) Microsoft Corporation. All rights reserved.
This code is licensed under the Visual Studio SDK license terms.
THIS CODE IS PROVIDED *AS IS* WITHOUT WARRANTY OF
ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING ANY
IMPLIED WARRANTIES OF FITNESS FOR A PARTICULAR
PURPOSE, MERCHANTABILITY, OR NON-INFRINGEMENT.

***************************************************************************/

#include "project.h"

#include "resource.h"       //for IDB_COMPLETION_SET
#include "dllmain.h"        //for g_instance
#include "completionset.h"
#include "source.h"         //for IsCommitChar

#include <strsafe.h>

/*---------------------------------------------------------
  CompletionSet 
-----------------------------------------------------------*/
CompletionSet::CompletionSet( in HANDLE imageList, in long glyphCount, in Source* source )
{
  TRACE_CREATE( "CompletionSet", static_cast<IVsCompletionSet*>(this) );
  ASSERT(source);

  m_refCount      = 1;
  m_imageList     = imageList;
  m_glyphCount    = glyphCount;
  m_source        = source;     ADDREF(source);

  m_textView      = NULL;
  m_decls         = NULL;
  m_displayed     = false;
  m_completeWord  = false;
}


CompletionSet::~CompletionSet()
{
  TRACE_DESTROY( static_cast<IVsCompletionSet*>(this) );
  Done();
  RELEASE(m_source);
}


STDMETHODIMP CompletionSet::Init( in IVsTextView* textView, in IDeclarations* decls, in bool completeWord )
{
//  TRACE("CompletionSet::Init");
  REFARG(textView);
  REFARG(decls);
  HRESULT hr;
  
  Done();
  
  ASSERT(!m_displayed);       m_displayed = false;
  ASSERT(m_textView == NULL); RELEASE(m_textView);
  ASSERT(m_decls == NULL);    RELEASE(m_decls);

  //check if we have members
  long count = 0;
  hr = decls->GetCount( &count );
  if (FAILED(hr)) return hr;
  if (count <= 0) return S_FALSE;

  //initialise and refresh
  m_decls        = decls;    ADDREF(m_decls);
  m_textView     = textView; ADDREF(m_textView);  
  m_completeWord = completeWord;

  DWORD flags = UCS_NAMESCHANGED;
  if (m_completeWord) flags |= UCS_COMPLETEWORD;

  hr = m_textView->UpdateCompletionStatus( this, flags );
  if (FAILED(hr)) return hr;

  m_displayed = true;
  return S_OK;
}


STDMETHODIMP CompletionSet::Done()
{
//  TRACE("CompletionSet::Done");
  if (m_displayed && m_textView)
  {
    m_textView->UpdateCompletionStatus( NULL, 0 );
    ASSERT(!m_displayed);
  }

  m_displayed = false;
  RELEASE(m_textView);
  RELEASE(m_decls);
  return S_OK;
}

/*---------------------------------------------------------
  IUnknown
-----------------------------------------------------------*/
STDMETHODIMP CompletionSet::QueryInterface( in REFIID iid, out void** obj )
{
  OUTARG(obj);

  if (iid == IID_IUnknown || iid == IID_IVsCompletionSet)
  {
    TRACE("CompletionSet::QueryInterface for IUnknown/IVsCompletionSet");
    *obj = static_cast<IVsCompletionSet*>(this);
  } 
  else
    return E_NOINTERFACE;

  AddRef();
  return S_OK;
}

STDMETHODIMP_(ULONG) CompletionSet::AddRef()
{
  return IncRefCount(&m_refCount);
}

STDMETHODIMP_(ULONG) CompletionSet::Release()
{
  if (DecRefCount(&m_refCount) == 0)
  {
    delete this;
    return 0;
  }
  else
    return m_refCount;
}


/*---------------------------------------------------------
  IVsCompletionSet
-----------------------------------------------------------*/
STDMETHODIMP_(DWORD) CompletionSet::GetFlags( void )
{
  DWORD flags = CSF_HAVEDESCRIPTIONS;

  if (m_source)  flags |= (CSF_CUSTOMCOMMIT | CSF_INITIALEXTENTKNOWN);

  if (m_decls)
  {
    HRESULT      hr;
    long         index;
    VARIANT_BOOL unique;
    hr = m_decls->GetBestMatch( NULL, &index, &unique );
    if (hr != E_NOTIMPL) flags |= CSF_CUSTOMMATCHING;
  }

  return flags;
}

STDMETHODIMP_(long)  CompletionSet::GetCount( void )
{
  //TRACE("CompletionSet::GetCount");
  if (!m_decls) return 0;

  HRESULT hr;
  long    count = 0;
  hr = m_decls->GetCount(&count );
  if (FAILED(hr)) return 0;

  return count;
}
  

STDMETHODIMP CompletionSet::GetImageList( out HANDLE* imageList )
{
  OUTARG(imageList);
  if (!m_imageList) return E_NOTIMPL;

  *imageList = m_imageList;
  return S_OK;
}

STDMETHODIMP CompletionSet::GetDisplayText( in long index, out const WCHAR** text, out long* glyph )
{
  OUTARG(text);
  if (glyph) *glyph = 0;
  ASSERT(m_decls); if (!m_decls) return E_UNEXPECTED;

  HRESULT hr;
  if (glyph) 
  {
    hr = m_decls->GetGlyph( index, glyph );
    if (*glyph >= m_glyphCount) *glyph = 0;
    if (FAILED(hr)) return hr;
  }

  BSTR bstrText;
  hr = m_decls->GetName( index, &bstrText );
  if (FAILED(hr)) return hr;
  *text = bstrText;
  
  return S_OK;
}

STDMETHODIMP CompletionSet::GetDescriptionText( in long index, out BSTR* description )
{
  OUTARG(description);
  ASSERT(m_decls); if (!m_decls) return E_UNEXPECTED;

  HRESULT hr;
  BSTR bstrDesc;
  hr = m_decls->GetDescription( index, &bstrDesc );
  if (FAILED(hr)) return hr;
  *description = ::SysAllocString(bstrDesc);

  return S_OK;
}



#ifdef VS6
STDMETHODIMP CompletionSet::GetInitialExtent( out long* line, 
                                              out long* startIdx, 
                                              out long* len )
{
  //TRACE("CompletionSet::GetInitialExtent");
  OUTARG(len);
  OUTARG(startIdx);
  OUTARG(line);
  ASSERT(m_textView); if (!m_textView) return E_UNEXPECTED;
  
  long idx;
  HRESULT hr = m_textView->GetCaretPos( line, &idx );
  *len = 0;
  *startIdx = idx;
  if (FAILED(hr)) return hr;

  long endIdx;
  hr = m_source->GetWordExtent( *line, idx, WORDEXT_CURRENT, startIdx, &endIdx );
  if (FAILED(hr)) return hr;
  if (hr == S_FALSE && idx > 0) 
  {
    hr = m_source->GetWordExtent( *line, idx-1, WORDEXT_CURRENT, startIdx, &endIdx );
    if (FAILED(hr)) return hr;
    if (hr == S_FALSE) { *startIdx = idx; return hr; }
  }
  *len = endIdx - *startIdx;

  return S_OK;
}
#else
STDMETHODIMP CompletionSet::GetInitialExtent( out long* line, out long* startIdx, out long* endIdx )
{
  //TRACE("CompletionSet::GetInitialExtent");
  OUTARG(endIdx);
  OUTARG(startIdx);
  OUTARG(line);
  ASSERT(m_textView); if (!m_textView) return E_UNEXPECTED;
  
  long idx;
  HRESULT hr = m_textView->GetCaretPos( line, &idx );
  *endIdx = *startIdx = idx;
  if (FAILED(hr)) return hr;

  hr = m_source->GetWordExtent( *line, idx, WORDEXT_CURRENT, startIdx, endIdx );
  if (FAILED(hr)) return hr;
  if (hr == S_FALSE && idx > 0) 
  {
    hr = m_source->GetWordExtent( *line, idx-1, WORDEXT_CURRENT, startIdx, endIdx );
    if (FAILED(hr)) return hr;
    if (hr == S_FALSE) { *endIdx = *startIdx = idx; return hr; }
  }
  
  return S_OK;
}
#endif

STDMETHODIMP CompletionSet::GetBestMatch( in const WCHAR *textSoFar, in long len, out long* index, out DWORD* flags )
{
  TRACE2( "CompletionSet::GetBestMatch: len=%i, text=%S", len, textSoFar );
  OUTARG( flags );
  OUTARG( index );

  VARIANT_BOOL uniqueMatch = VARIANT_FALSE;
  BSTR         prefix      = NULL;   bstrCopy( prefix, textSoFar, len );
  HRESULT hr = m_decls->GetBestMatch( prefix, index, &uniqueMatch );
  bstrFree(prefix);
  if (FAILED(hr)) return hr;

  if (hr == S_OK)  *flags |= GBM_SELECT;
  if (uniqueMatch) *flags |= GBM_UNIQUE;

  return S_OK;
}
    
STDMETHODIMP_(void) CompletionSet::Dismiss()
{
//  TRACE("CompletionSet::Dismiss");
//  ASSERT(m_displayed);

  m_displayed = false;
  RELEASE(m_decls);
  RELEASE(m_textView);
}

STDMETHODIMP CompletionSet::OnCommit( in const WCHAR* textSoFar, in long index
					                , in BOOL selected, in WCHAR commitChar, out BSTR* completeWord )
{
  TRACE2("CompletionSet::OnCommit: %S, %c", textSoFar, commitChar );
  OUTARG(completeWord);
  REFARG(textSoFar);

  HRESULT hr = S_FALSE;
  bool fSpecialChar = (0 == commitChar) || (VK_RETURN == commitChar) || (VK_TAB == commitChar);

  // Check if there is any match for the current text
  long matchIndex = -1;
  VARIANT_BOOL uniqueMatch = VARIANT_FALSE;
  size_t lenSoFar = lstrlenW(textSoFar);

  if (selected)
  {
    if ( fSpecialChar )
    {
      // Commit char zero means that the user has selected the entry with index "index"
      // from the list, so we have to assume that the match is found and that it is unique.
      uniqueMatch = true;
      matchIndex = index;
      hr = S_OK;
    }
    else if (m_decls)
    {
      // Try to find the token in the list of the declarations.
      // The string to check is the one built so far plus the commit char.
      size_t len = lenSoFar + 1;
      CComBSTR prefix(len);
      if (prefix == NULL) return E_OUTOFMEMORY;
      hr = StringCchCopy(prefix, len, textSoFar);
      if (FAILED(hr)) return hr;
      prefix[len-1] = commitChar;
      prefix[len] = 0;
   
      hr = m_decls->GetBestMatch( prefix, &matchIndex, &uniqueMatch );
      if (FAILED(hr)) return hr;
    }
  }

  // If the token is not found or is not unique this operation can not be
  // considered a commit, so return S_FALSE.
  if ((S_FALSE==hr) || (VARIANT_FALSE==uniqueMatch))
  {
      return S_FALSE;
  }

  // Commit the operation.
  const WCHAR* displayText = NULL;
  hr = GetDisplayText( matchIndex, &displayText, NULL );
  if (FAILED(hr)) return hr;

  size_t displayLen = lstrlenW(displayText);
  if ( fSpecialChar || (displayLen == lenSoFar+1))
  {
    size_t wordLen = fSpecialChar ? displayLen : lenSoFar;
    *completeWord = SysAllocStringLen(displayText, wordLen);
    if (NULL == *completeWord)
        return E_OUTOFMEMORY;
    return S_OK;
  }


  return S_FALSE;
}
